package shape

import kotlinx.io.core.ByteOrder
import kotlinx.io.core.ByteReadPacket

/**
 * Reads 36 bytes
 * 0-3:   header
 * 4-23:  reserved
 * 24-27: version
 * 28-31: fileLength
 * 32-35: shapeId
 */
internal fun readShpHeader(packet: ByteReadPacket): ShpHeader
{
    packet.byteOrder = ByteOrder.BIG_ENDIAN

    val headerId = packet.readInt()

    // skip empty
    packet.discard(20)

    val fileLength = packet.readInt()

    packet.byteOrder = ByteOrder.LITTLE_ENDIAN

    val version = packet.readInt()

    val shapeTypeId = packet.readInt()

    val shapeType = when (shapeTypeId)
    {
        0    -> ShapeType.NULL
        1    -> ShapeType.POINT
        3    -> ShapeType.POLYLINE
        5    -> ShapeType.POLYGON
        8    -> ShapeType.MULTIPOINT
        11   -> ShapeType.POINTZ
        13   -> ShapeType.POLYLINEZ
        15   -> ShapeType.POLYGONZ
        18   -> ShapeType.MULTIPOINTZ
        21   -> ShapeType.POINTM
        23   -> ShapeType.POLYLINEM
        25   -> ShapeType.POLYGONM
        28   -> ShapeType.MULTIPOINTM
        31   -> ShapeType.MULTIPATCH
        else -> throw IllegalStateException("cannot process type $shapeTypeId")
    }

    return ShpHeader(headerId, fileLength, version, shapeType)
}

/**
 * Reads 36 bytes
 * 0-3:   header
 * 4-23:  reserved
 * 24-27: version
 * 28-31: fileLength
 * 32-35: shapeId
 */
internal fun readShpHeaderCorrect(packet: ByteReadPacket): ShpHeader
{
    val headerId = packet.readInt()

    // skip empty
    packet.discard(20)

    val fileLength = packet.readInt()
    val version = packet.readInt().swapEndianness()
    val shapeTypeId = packet.readInt().swapEndianness()

    val shapeType = when (shapeTypeId)
    {
        0    -> ShapeType.NULL
        1    -> ShapeType.POINT
        3    -> ShapeType.POLYLINE
        5    -> ShapeType.POLYGON
        8    -> ShapeType.MULTIPOINT
        11   -> ShapeType.POINTZ
        13   -> ShapeType.POLYLINEZ
        15   -> ShapeType.POLYGONZ
        18   -> ShapeType.MULTIPOINTZ
        21   -> ShapeType.POINTM
        23   -> ShapeType.POLYLINEM
        25   -> ShapeType.POLYGONM
        28   -> ShapeType.MULTIPOINTM
        31   -> ShapeType.MULTIPATCH
        else -> throw IllegalStateException("cannot process type $shapeTypeId")
    }

    return ShpHeader(headerId, fileLength, version, shapeType)
}

fun Int.swapEndianness() = swapInteger(this)

private fun swapInteger(value: Int): Int
{
    return (value shr 0 and 0xff shl 24) +
            (value shr 8 and 0xff shl 16) +
            (value shr 16 and 0xff shl 8) +
            (value shr 24 and 0xff shl 0)
}